@@ -3,6 +3,7 @@
     Copyright (C) 2008-2010 Dario Freddi <drf@kde.org>
     Copyright (C) 2010 Alejandro Fiestas <alex@eyeos.org>
     Copyright (C) 2010-2013 Lukáš Tinkl <ltinkl@redhat.com>
+    Copyright (C) 2015 Kai Uwe Broulik <kde@privat.broulik.de>
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Library General Public
@@ -23,10 +24,12 @@
 #include "powerdevilupowerbackend.h"
 
 #include <powerdevil_debug.h>
+#include <PowerDevilSettings.h>
 
 #include <qtextstream.h>
 #include <QtDBus>
 #include <QDebug>
+#include <QPropertyAnimation>
 
 #include <kauthexecutejob.h>
 #include <KPluginFactory>
@@ -43,7 +46,8 @@
 
 PowerDevilUPowerBackend::PowerDevilUPowerBackend(QObject* parent)
     : BackendInterface(parent),
-      m_brightnessControl(0),
+      m_displayDevice(Q_NULLPTR),
+      m_brightnessControl(Q_NULLPTR),
       m_kbdMaxBrightness(0),
       m_lidIsPresent(false), m_lidIsClosed(false), m_onBattery(false)
 {
@@ -134,21 +138,21 @@ void PowerDevilUPowerBackend::init()
     if (!m_brightnessControl->isSupported()) {
         qCDebug(POWERDEVIL) << "Falling back to helper to get brightness";
 
-        KAuth::Action brightnessValueAction("org.kde.powerdevil.backlighthelper.brightnessvalue");
-        brightnessValueAction.setHelperId(HELPER_ID);
-        KAuth::ExecuteJob *brightnessValueJob = brightnessValueAction.execute();
-        if (!brightnessValueJob->exec()) {
-            qCWarning(POWERDEVIL) << "org.kde.powerdevil.backlighthelper.brightnessvalue failed";
+        KAuth::Action brightnessAction("org.kde.powerdevil.backlighthelper.brightness");
+        brightnessAction.setHelperId(HELPER_ID);
+        KAuth::ExecuteJob *brightnessJob = brightnessAction.execute();
+        if (!brightnessJob->exec()) {
+            qCWarning(POWERDEVIL) << "org.kde.powerdevil.backlighthelper.brightness failed";
         } else {
-            m_cachedBrightnessMap.insert(Screen, brightnessValueJob->data()["brightnessvalue"].toFloat());
+            m_cachedBrightnessMap.insert(Screen, brightnessJob->data()["brightness"].toFloat());
 
-            KAuth::Action brightnessValueMaxAction("org.kde.powerdevil.backlighthelper.brightnessvaluemax");
-            brightnessValueMaxAction.setHelperId(HELPER_ID);
-            KAuth::ExecuteJob *brightnessValueMaxJob = brightnessValueMaxAction.execute();
-            if (brightnessValueMaxJob->exec()) {
-                m_brightnessValueMax = brightnessValueMaxJob->data()["brightnessvaluemax"].toInt();
+            KAuth::Action brightnessMaxAction("org.kde.powerdevil.backlighthelper.brightnessmax");
+            brightnessMaxAction.setHelperId(HELPER_ID);
+            KAuth::ExecuteJob *brightnessMaxJob = brightnessMaxAction.execute();
+            if (brightnessMaxJob->exec()) {
+                m_brightnessMax = brightnessMaxJob->data()["brightnessmax"].toInt();
             } else {
-                qCWarning(POWERDEVIL) << "org.kde.powerdevil.backlighthelper.brightnessvaluemax failed";
+                qCWarning(POWERDEVIL) << "org.kde.powerdevil.backlighthelper.brightnessmax failed";
             }
 
             KAuth::Action syspathAction("org.kde.powerdevil.backlighthelper.syspath");
@@ -168,9 +172,18 @@ void PowerDevilUPowerBackend::init()
         qCDebug(POWERDEVIL) << "Using XRandR";
         m_randrHelper = XRandRXCBHelper::self();
         Q_ASSERT(m_randrHelper);
-        connect(m_randrHelper, SIGNAL(brightnessChanged()), this, SLOT(slotScreenBrightnessChanged()));
-        m_cachedBrightnessMap.insert(Screen, brightnessValue(Screen));
+        connect(m_randrHelper, &XRandRXCBHelper::brightnessChanged, this, &PowerDevilUPowerBackend::slotScreenBrightnessChanged);
+        m_cachedBrightnessMap.insert(Screen, brightness(Screen));
         screenBrightnessAvailable = true;
+
+        const int duration = PowerDevilSettings::brightnessAnimationDuration();
+        if (duration > 0 && brightnessMax() >= PowerDevilSettings::brightnessAnimationThreshold()) {
+            m_brightnessAnimation = new QPropertyAnimation(this);
+            m_brightnessAnimation->setTargetObject(this);
+            m_brightnessAnimation->setDuration(duration);
+            m_brightnessAnimation->setEasingCurve(QEasingCurve::InOutQuad);
+            connect(m_brightnessAnimation, &QPropertyAnimation::valueChanged, this, &PowerDevilUPowerBackend::animationValueChanged);
+        }
     }
 
     // Capabilities
@@ -213,7 +226,7 @@ void PowerDevilUPowerBackend::init()
         // TODO Do a proper check if the kbd backlight dbus object exists. But that should work for now ..
         if (m_kbdMaxBrightness) {
             controls.insert(QLatin1String("KBD"), Keyboard);
-            m_cachedBrightnessMap.insert(Keyboard, brightnessValue(Keyboard));
+            m_cachedBrightnessMap.insert(Keyboard, brightness(Keyboard));
             qCDebug(POWERDEVIL) << "current keyboard backlight brightness value: " << m_cachedBrightnessMap.value(Keyboard);
             connect(m_kbdBacklight, SIGNAL(BrightnessChanged(int)), this, SLOT(onKeyboardBrightnessChanged(int)));
         }
@@ -256,24 +269,12 @@ void PowerDevilUPowerBackend::init()
     }
 
     // battery
-    QList<RecallNotice> recallList;
     foreach(OrgFreedesktopUPowerDeviceInterface * upowerDevice, m_devices) {
         if (upowerDevice->type() == 2 && upowerDevice->powerSupply()) {
             QString udi = upowerDevice->path();
             setCapacityForBattery(udi, qRound(upowerDevice->capacity()));  // acknowledge capacity
-
-            if (upowerDevice->recallNotice()) {                            // check for recall notices
-                RecallNotice notice;
-                notice.batteryId = udi;
-                notice.url = upowerDevice->recallUrl();
-                notice.vendor = upowerDevice->recallVendor();
-
-                recallList.append(notice);
-            }
         }
     }
-    if (!recallList.isEmpty())
-        setRecallNotices(recallList);
 
     // backend ready
     setBackendIsReady(controls, supported);
@@ -307,33 +308,37 @@ int PowerDevilUPowerBackend::brightnessKeyPressed(PowerDevil::BrightnessLogic::B
         return -1; // ignore as we are not able to determine the brightness level
     }
 
-    int currentBrightness = brightnessValue(controlType);
-    if (currentBrightness !=  m_cachedBrightnessMap.value(controlType)) {
+    int currentBrightness = brightness(controlType);
+    if (currentBrightness != m_cachedBrightnessMap.value(controlType)) {
         m_cachedBrightnessMap[controlType] = currentBrightness;
         return currentBrightness;
     }
 
-    int maxBrightness = brightnessValueMax(controlType),
-        newBrightness = calculateNextStep(currentBrightness, maxBrightness, controlType, type);
+    int maxBrightness = brightnessMax(controlType);
+    int newBrightness = calculateNextStep(currentBrightness, maxBrightness, controlType, type);
 
     if (newBrightness < 0) {
         return -1;
     }
 
-    setBrightnessValue(newBrightness, controlType);
+    setBrightness(newBrightness, controlType);
     return newBrightness;
 }
 
-int PowerDevilUPowerBackend::brightnessValue(PowerDevil::BackendInterface::BrightnessControlType type) const
+int PowerDevilUPowerBackend::brightness(PowerDevil::BackendInterface::BrightnessControlType type) const
 {
     int result = 0;
 
     if (type == Screen) {
         if (m_brightnessControl->isSupported()) {
-            //qCDebug(POWERDEVIL) << "Calling xrandr brightness";
-            result = (int) m_brightnessControl->brightnessValue();
+            if (m_brightnessAnimation && m_brightnessAnimation->state() == QPropertyAnimation::Running) {
+                result = m_brightnessAnimation->endValue().toInt();
+            } else {
+                //qCDebug(POWERDEVIL) << "Calling xrandr brightness";
+                result = (int) m_brightnessControl->brightness();
+            }
         } else {
-            return m_cachedBrightnessMap[Screen];
+            result = m_cachedBrightnessMap[Screen];
         }
         qCDebug(POWERDEVIL) << "Screen brightness value: " << result;
     } else if (type == Keyboard) {
@@ -344,16 +349,16 @@ int PowerDevilUPowerBackend::brightnessValue(PowerDevil::BackendInterface::Brigh
     return result;
 }
 
-int PowerDevilUPowerBackend::brightnessValueMax(PowerDevil::BackendInterface::BrightnessControlType type) const
+int PowerDevilUPowerBackend::brightnessMax(PowerDevil::BackendInterface::BrightnessControlType type) const
 {
     int result = 0;
 
     if (type == Screen) {
         if (m_brightnessControl->isSupported()) {
             //qCDebug(POWERDEVIL) << "Calling xrandr brightness";
-            result = (int) m_brightnessControl->brightnessValueMax();
+            result = (int) m_brightnessControl->brightnessMax();
         } else {
-            return m_brightnessValueMax;
+            result = m_brightnessMax;
         }
         qCDebug(POWERDEVIL) << "Screen brightness value max: " << result;
     } else if (type == Keyboard) {
@@ -364,40 +369,45 @@ int PowerDevilUPowerBackend::brightnessValueMax(PowerDevil::BackendInterface::Br
     return result;
 }
 
-bool PowerDevilUPowerBackend::setBrightnessValue(int brightnessValue, PowerDevil::BackendInterface::BrightnessControlType type)
+void PowerDevilUPowerBackend::setBrightness(int value, PowerDevil::BackendInterface::BrightnessControlType type)
 {
-    bool success = false;
     if (type == Screen) {
-        qCDebug(POWERDEVIL) << "set screen brightness value: " << brightnessValue;
+        qCDebug(POWERDEVIL) << "set screen brightness value: " << value;
         if (m_brightnessControl->isSupported()) {
-            m_brightnessControl->setBrightnessValue(brightnessValue);
+            if (m_brightnessAnimation) {
+                m_brightnessAnimation->stop();
+                m_brightnessAnimation->setStartValue(brightness());
+                m_brightnessAnimation->setEndValue(value);
+                m_brightnessAnimation->start();
+            } else {
+                m_brightnessControl->setBrightness(value);
+            }
         } else {
             //qCDebug(POWERDEVIL) << "Falling back to helper to set brightness";
-            KAuth::Action action("org.kde.powerdevil.backlighthelper.setbrightnessvalue");
+            KAuth::Action action("org.kde.powerdevil.backlighthelper.setbrightness");
             action.setHelperId(HELPER_ID);
-            action.addArgument("brightnessvalue", brightnessValue);
+            action.addArgument("brightness", value);
             KAuth::ExecuteJob *job = action.execute();
             // we don't care about the result since executing the job sync is bad
             job->start();
         }
-
-        success = true;
     } else if (type == Keyboard) {
-        qCDebug(POWERDEVIL) << "set kbd backlight value: " << brightnessValue;
-        m_kbdBacklight->SetBrightness(brightnessValue);
-        success = true;
+        qCDebug(POWERDEVIL) << "set kbd backlight value: " << value;
+        m_kbdBacklight->SetBrightness(value);
     }
-
-    return success;
 }
 
 void PowerDevilUPowerBackend::slotScreenBrightnessChanged()
 {
-    int value = brightnessValue(Screen);
+    if (m_brightnessAnimation && m_brightnessAnimation->state() != QPropertyAnimation::Stopped) {
+        return;
+    }
+
+    int value = brightness(Screen);
     qCDebug(POWERDEVIL) << "Brightness changed!!";
     if (value != m_cachedBrightnessMap[Screen]) {
         m_cachedBrightnessMap[Screen] = value;
-        onBrightnessChanged(Screen, value, brightnessValueMax(Screen));
+        onBrightnessChanged(Screen, value, brightnessMax(Screen));
     }
 }
 
@@ -406,7 +416,7 @@ void PowerDevilUPowerBackend::onKeyboardBrightnessChanged(int value)
     qCDebug(POWERDEVIL) << "Keyboard brightness changed!!";
     if (value != m_cachedBrightnessMap[Keyboard]) {
         m_cachedBrightnessMap[Keyboard] = value;
-        onBrightnessChanged(Keyboard, value, brightnessValueMax(Keyboard));
+        onBrightnessChanged(Keyboard, value, brightnessMax(Keyboard));
     }
 }
 
@@ -431,6 +441,14 @@ void PowerDevilUPowerBackend::enumerateDevices()
         addDevice(device.path());
     }
 
+    QDBusReply<QDBusObjectPath> reply = m_upowerInterface->call("GetDisplayDevice");
+    if (reply.isValid()) {
+        const QString path = reply.value().path();
+        if (!path.isEmpty() && path != QStringLiteral("/")) {
+            m_displayDevice = new OrgFreedesktopUPowerDeviceInterface(UPOWER_SERVICE, path, QDBusConnection::systemBus(), this);
+        }
+    }
+
     updateDeviceProps();
 
     if (m_onBattery)
@@ -484,14 +502,48 @@ void PowerDevilUPowerBackend::updateDeviceProps()
 {
     qlonglong remainingTime = 0;
 
-    foreach(OrgFreedesktopUPowerDeviceInterface * upowerDevice, m_devices) {
-        const uint type = upowerDevice->type();
-        if (( type == 2 || type == 3) && upowerDevice->powerSupply()) {
-            const uint state = upowerDevice->state();
-            if (state == 1) // charging
-                remainingTime += upowerDevice->timeToFull();
-            else if (state == 2) //discharging
-                remainingTime += upowerDevice->timeToEmpty();
+    if (m_displayDevice && m_displayDevice->isPresent()) {
+        const uint state = m_displayDevice->state();
+        if (state == 1) // charging
+            remainingTime = m_displayDevice->timeToFull();
+        else if (state == 2) //discharging
+            remainingTime = m_displayDevice->timeToEmpty();
+    } else {
+        qreal energyTotal = 0.0;
+        qreal energyRateTotal = 0.0;
+        qreal energyFullTotal = 0.0;
+        uint stateTotal = 0;
+
+        foreach(OrgFreedesktopUPowerDeviceInterface * upowerDevice, m_devices) {
+            const uint type = upowerDevice->type();
+            if (( type == 2 || type == 3) && upowerDevice->powerSupply()) {
+                const uint state = upowerDevice->state();
+                energyFullTotal += upowerDevice->energyFull();
+                energyTotal += upowerDevice->energy();
+                energyRateTotal += upowerDevice->energyRate();
+
+                if (state == 1) { // total is charging
+                    stateTotal = 1;
+                } else if (state == 2 && stateTotal != 1) { // total is discharging
+                    stateTotal = 2;
+                } else if (state == 4 && stateTotal != 0) { // total is fully-charged
+                    stateTotal = 4;
+                }
+
+                if (state == 1) { // charging
+                    remainingTime += upowerDevice->timeToFull();
+                } else if (state == 2) { // discharging
+                    remainingTime += upowerDevice->timeToEmpty();
+                }
+            }
+        }
+
+        if (energyRateTotal > 0) {
+            if (stateTotal == 1) { // charging
+                remainingTime = 3600 * ((energyFullTotal - energyTotal) / energyRateTotal);
+            } else if (stateTotal == 2) { // discharging
+                remainingTime = 3600 * (energyTotal / energyRateTotal);
+            }
         }
     }
 
@@ -551,4 +603,7 @@ void PowerDevilUPowerBackend::slotLogin1Resuming(bool active)
     }
 }
 
-#include "powerdevilupowerbackend.moc"
+void PowerDevilUPowerBackend::animationValueChanged(const QVariant &value)
+{
+    m_brightnessControl->setBrightness(value.toInt());
+}
